import os
import time
from collections import OrderedDict

import pytest
from ingenialink.canopen.network import CanopenNetwork
from ingenialink.canopen.servo import CanopenServo
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.exceptions import ILError
from ingenialink.network import SlaveInfo
from ingenialink.servo import SERVO_STATE

from ingeniamotion import MotionController
from ingeniamotion.enums import CAN_BAUDRATE, CAN_DEVICE
from ingeniamotion.exceptions import IMException, IMRegisterNotExist, IMRegisterWrongAccess

TEST_ENSEMBLE_FW_FILE = "tests/resources/example_ensemble_fw.zfu"


@pytest.mark.virtual
def test_connect_servo_eoe(read_config):
    mc = MotionController()
    assert "eoe_test" not in mc.servos
    assert "eoe_test" not in mc.net
    mc.communication.connect_servo_eoe(
        read_config["ip"], read_config["dictionary"], alias="eoe_test"
    )
    assert "eoe_test" in mc.servos and mc.servos["eoe_test"] is not None
    assert "eoe_test" in mc.net and mc.net["eoe_test"] is not None


@pytest.mark.virtual
def test_connect_servo_eoe_no_dictionary_error(read_config):
    mc = MotionController()
    with pytest.raises(FileNotFoundError):
        mc.communication.connect_servo_eoe(read_config["ip"], "no_dictionary", alias="eoe_test")


@pytest.mark.virtual
def test_connect_servo_ethernet(read_config):
    mc = MotionController()
    assert "eoe_test" not in mc.servos
    assert "eoe_test" not in mc.net
    mc.communication.connect_servo_ethernet(
        read_config["ip"], read_config["dictionary"], alias="eoe_test"
    )
    assert "eoe_test" in mc.servos and mc.servos["eoe_test"] is not None
    assert "eoe_test" in mc.net and mc.net["eoe_test"] is not None


@pytest.mark.virtual
def test_connect_servo_ethernet_no_dictionary_error(read_config):
    mc = MotionController()
    with pytest.raises(FileNotFoundError):
        mc.communication.connect_servo_ethernet(
            read_config["ip"], "no_dictionary", alias="eoe_test"
        )


@pytest.mark.smoke
@pytest.mark.eoe
@pytest.mark.parametrize(
    "coco_dict_path",
    [
        True,
        False,
    ],
)
def test_connect_servo_comkit_no_dictionary_error(coco_dict_path, read_config):
    mc = MotionController()
    if coco_dict_path:
        coco_dict_path = read_config["dictionary"]
        moco_dict_path = "no_dictionary"
    else:
        coco_dict_path = "no_dictionary"
        moco_dict_path = read_config["dictionary"]
    with pytest.raises(FileNotFoundError):
        mc.communication.connect_servo_comkit(
            read_config["ip"], coco_dict_path, moco_dict_path, alias="eoe_test"
        )


@pytest.mark.smoke
@pytest.mark.soem
def test_get_ifname_from_interface_ip(mocker):
    ip = type("IP", (object,), {"ip": "192.168.2.1", "is_IPv4": True})
    adapter = type(
        "Adapter", (object,), {"ips": [ip], "name": b"{192D1D2F-C684-467D-A637-EC07BD434A63}"}
    )
    mocker.patch("ifaddr.get_adapters", return_value=[adapter])
    mc = MotionController()
    ifname = mc.communication.get_ifname_from_interface_ip("192.168.2.1")
    assert ifname == "\\Device\\NPF_{192D1D2F-C684-467D-A637-EC07BD434A63}"


@pytest.mark.skip(reason='This test enters in conflict with "disable_motor_fixture"')
@pytest.mark.smoke
@pytest.mark.canopen
def test_connect_servo_canopen(read_config):
    mc = MotionController()
    assert "canopen_test" not in mc.servos
    assert "canopen_test" not in mc.net
    device = CAN_DEVICE(read_config["device"])
    baudrate = CAN_BAUDRATE(read_config["baudrate"])
    mc.communication.connect_servo_canopen(
        device,
        read_config["dictionary"],
        read_config["node_id"],
        baudrate,
        read_config["channel"],
        alias="canopen_test",
    )
    assert "canopen_test" in mc.servos and mc.servos["canopen_test"] is not None
    assert "canopen_test" in mc.net and mc.net["canopen_test"] is not None
    mc.net["canopen_test"].disconnect()


@pytest.mark.smoke
@pytest.mark.canopen
@pytest.mark.skip
def test_connect_servo_canopen_busy_drive_error(motion_controller, read_config):
    mc, alias = motion_controller
    assert "canopen_test" not in mc.servos
    assert "canopen_test" not in mc.servo_net
    assert alias in mc.servos
    assert alias in mc.servo_net
    assert mc.servo_net[alias] in mc.net
    device = CAN_DEVICE(read_config["device"])
    baudrate = CAN_BAUDRATE(read_config["baudrate"])
    with pytest.raises(ILError):
        mc.communication.connect_servo_canopen(
            device,
            read_config["dictionary"],
            read_config["node_id"],
            baudrate,
            read_config["channel"],
            alias="canopen_test",
        )


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value",
    [
        ("CL_VOL_Q_SET_POINT", 0.34),
        ("CL_POS_SET_POINT_VALUE", -923),
        ("PROF_POS_OPTION_CODE", 1),
    ],
)
def test_get_register(motion_controller, uid, value):
    mc, alias = motion_controller
    drive = mc.servos[alias]
    drive.write(uid, value)
    test_value = mc.communication.get_register(uid, servo=alias)
    assert pytest.approx(test_value) == value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_register_wrong_uid(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(IMRegisterNotExist):
        mc.communication.get_register("WRONG_UID", servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value",
    [
        ("CL_VOL_Q_SET_POINT", -234),
        ("CL_POS_SET_POINT_VALUE", 23),
        ("PROF_POS_OPTION_CODE", 54),
    ],
)
def test_set_register(motion_controller, uid, value):
    mc, alias = motion_controller
    drive = mc.servos[alias]
    mc.communication.set_register(uid, value, servo=alias)
    test_value = drive.read(uid)
    assert pytest.approx(test_value) == value


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_register_wrong_uid(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(IMRegisterNotExist):
        mc.communication.set_register("WRONG_UID", 2, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value, fail",
    [
        ("CL_VOL_Q_SET_POINT", -234, False),
        ("CL_VOL_Q_SET_POINT", "I'm not a number", True),
        ("CL_VOL_Q_SET_POINT", 234.4, False),
        ("CL_POS_SET_POINT_VALUE", 1245, False),
        ("CL_POS_SET_POINT_VALUE", -1245, False),
        ("CL_POS_SET_POINT_VALUE", 1245.5421, True),
        ("PROF_POS_OPTION_CODE", -54, True),
        ("PROF_POS_OPTION_CODE", 54, False),
        ("PROF_POS_OPTION_CODE", "54", True),
    ],
)
def test_set_register_wrong_value_type(motion_controller, uid, value, fail):
    mc, alias = motion_controller
    if fail:
        with pytest.raises(TypeError):
            mc.communication.set_register(uid, value, servo=alias)
    else:
        mc.communication.set_register(uid, value, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_register_wrong_access(motion_controller):
    mc, alias = motion_controller
    uid = "DRV_STATE_STATUS"
    value = 0
    with pytest.raises(IMRegisterWrongAccess):
        mc.communication.set_register(uid, value, servo=alias)


def dummy_callback(status, _, axis):
    pass


@pytest.mark.smoke
def test_subscribe_servo_status(mocker, motion_controller):
    mc, alias = motion_controller
    axis = 1
    patch_callback = mocker.patch("tests.test_communication.dummy_callback")
    mc.communication.subscribe_servo_status(patch_callback, alias)
    time.sleep(0.5)
    mc.motion.motor_enable(alias, axis)
    time.sleep(0.5)
    mc.motion.motor_disable(alias, axis)
    time.sleep(0.5)
    expected_status = [SERVO_STATE.RDY, SERVO_STATE.ENABLED, SERVO_STATE.DISABLED]
    for index, call in enumerate(patch_callback.call_args_list):
        assert call[0][0] == expected_status[index]
        assert call[0][2] == axis


@pytest.mark.virtual
def test_load_firmware_canopen_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.communication.load_firmware_canopen("fake_fw_file.lfu", servo=alias)


@pytest.mark.virtual
def test_boot_mode_and_load_firmware_ethernet_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc, "_get_network", return_value=EthercatNetwork("fake_interface_name"))
    with pytest.raises(ValueError):
        mc.communication.boot_mode_and_load_firmware_ethernet("fake_fw_file.lfu", servo=alias)


@pytest.mark.virtual
def test_load_firmware_moco_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc, "_get_network", return_value=EthercatNetwork("fake_interface_name"))
    with pytest.raises(ValueError):
        mc.communication.load_firmware_moco("fake_fw_file.lfu", servo=alias)


@pytest.mark.virtual
def test_connect_servo_virtual():
    mc = MotionController()
    mc.communication.connect_servo_virtual(port=1062)
    assert mc.communication._Communication__virtual_drive is not None
    mc.communication.disconnect()
    assert mc.communication._Communication__virtual_drive is None


@pytest.mark.virtual
def test_connect_servo_virtual_custom_dictionary(read_config):
    mc = MotionController()
    mc.communication.connect_servo_virtual(dict_path=read_config["dictionary"], port=1062)
    assert mc.communication._Communication__virtual_drive is not None
    mc.communication.disconnect()
    assert mc.communication._Communication__virtual_drive is None


@pytest.mark.virtual
def test_scan_servos_canopen_with_info(mocker):
    mc = MotionController()
    detected_slaves = OrderedDict({31: SlaveInfo(1234, 123), 32: SlaveInfo(1234, 123)})
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.scan_slaves_info", return_value=detected_slaves
    )
    assert mc.communication.scan_servos_canopen_with_info(CAN_DEVICE.KVASER) == detected_slaves


@pytest.mark.virtual
def test_scan_servos_canopen(mocker):
    mc = MotionController()
    detected_slaves = [31, 32]
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.scan_slaves", return_value=detected_slaves
    )
    assert mc.communication.scan_servos_canopen(CAN_DEVICE.KVASER) == detected_slaves


@pytest.mark.virtual
def test_scan_servos_ethercat_with_info(mocker):
    mc = MotionController()
    detected_slaves = OrderedDict({1: SlaveInfo(1234, 123), 2: SlaveInfo(1234, 123)})
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.scan_slaves_info",
        return_value=detected_slaves,
    )
    assert mc.communication.scan_servos_ethercat_with_info("") == detected_slaves


@pytest.mark.virtual
def test_scan_servos_ethercat(mocker):
    mc = MotionController()
    detected_slaves = [1, 2]
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.scan_slaves",
        return_value=detected_slaves,
    )
    assert mc.communication.scan_servos_ethercat("") == detected_slaves


@pytest.mark.virtual
def test_unzip_ensemble_fw_file():
    mc = MotionController()
    mapping = mc.communication._Communication__unzip_ensemble_fw_file(TEST_ENSEMBLE_FW_FILE)
    path = os.path.abspath("ensemble_temp")
    assert mapping == {
        0: (os.path.join(path, "cap-net-1-e_2.4.0.lfu"), 123456, 4660),
        1: (os.path.join(path, "cap-net-2-e_2.4.0.lfu"), 123456, 16781876),
    }


@pytest.mark.virtual
def test__check_ensemble():
    mc = MotionController()
    mapping = mc.communication._Communication__unzip_ensemble_fw_file(TEST_ENSEMBLE_FW_FILE)
    product_code = 123456
    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(product_code, 16781876),
            4: SlaveInfo(product_code, 4660),
            5: SlaveInfo(product_code, 16781876),
            7: SlaveInfo(654321, 1236),
        }
    )
    for slave_id in [1, 2]:
        assert mc.communication._Communication__check_ensemble(slaves, slave_id, mapping) == 1

    for slave_id in [4, 5]:
        assert mc.communication._Communication__check_ensemble(slaves, slave_id, mapping) == 4


@pytest.mark.virtual
def test__check_ensemble_wrong():
    mc = MotionController()
    mapping = mc.communication._Communication__unzip_ensemble_fw_file(TEST_ENSEMBLE_FW_FILE)
    product_code = 123456
    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(product_code, 16781876),
            3: SlaveInfo(654321, 1236),
        }
    )

    with pytest.raises(IMException) as exc_info:
        mc.communication._Communication__check_ensemble(slaves, 3, mapping)
    assert str(exc_info.value) == "The selected drive is not part of the ensemble."

    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(654321, 16781876),
        }
    )
    with pytest.raises(IMException) as exc_info:
        mc.communication._Communication__check_ensemble(slaves, 1, mapping)
    assert (
        str(exc_info.value)
        == "Wrong ensemble. The slave 2 has wrong product code or revision number."
    )

    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 16781876),
            2: SlaveInfo(product_code, 4660),
        }
    )
    with pytest.raises(IMException) as exc_info:
        mc.communication._Communication__check_ensemble(slaves, 2, mapping)
    assert str(exc_info.value) == "Wrong ensemble. The slave 2 is not detected."


@pytest.mark.virtual
@pytest.mark.parametrize("revision_number,expected_id_offset", [(4660, 0), (16781876, 1)])
def test_check_slave_in_ensemble(revision_number, expected_id_offset):
    mc = MotionController()
    mapping = mc.communication._Communication__unzip_ensemble_fw_file(TEST_ENSEMBLE_FW_FILE)
    product_code = 123456
    slave_info = SlaveInfo(product_code, revision_number)

    slave_id_offset = mc.communication._Communication__check_slave_in_ensemble(slave_info, mapping)

    assert slave_id_offset == expected_id_offset


@pytest.mark.virtual
def test_check_slave_in_ensemble_drive_not_in_ensemble():
    mc = MotionController()
    mapping = mc.communication._Communication__unzip_ensemble_fw_file(TEST_ENSEMBLE_FW_FILE)
    product_code = 654321
    revision_number = 4660
    slave_info = SlaveInfo(product_code, revision_number)

    with pytest.raises(IMException) as exc_info:
        mc.communication._Communication__check_slave_in_ensemble(slave_info, mapping)
    assert str(exc_info.value) == "The selected drive is not part of the ensemble."


@pytest.mark.virtual
def test_load_ensemble_fw_ecat(mocker):
    product_code = 123456
    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(product_code, 16781876),
            3: SlaveInfo(product_code, 4660),
            4: SlaveInfo(product_code, 16781876),
            5: SlaveInfo(654321, 1236),
        }
    )
    temp_path = os.path.abspath("ensemble_temp")
    fw_file1 = os.path.join(temp_path, "cap-net-1-e_2.4.0.lfu")
    fw_file2 = os.path.join(temp_path, "cap-net-2-e_2.4.0.lfu")
    mc = MotionController()
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.scan_slaves_info", return_value=slaves
    )
    for slave in [1, 2]:
        patch_fw_callback = mocker.patch(
            "ingenialink.ethercat.network.EthercatNetwork.load_firmware"
        )
        mc.communication.load_firmware_ecat("", TEST_ENSEMBLE_FW_FILE, slave=slave)
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0] == (fw_file1, False, 1)
        assert patch_fw_callback.call_args_list[1][0] == (fw_file2, False, 2)

    for slave in [3, 4]:
        patch_fw_callback = mocker.patch(
            "ingenialink.ethercat.network.EthercatNetwork.load_firmware"
        )
        mc.communication.load_firmware_ecat("", TEST_ENSEMBLE_FW_FILE, slave=slave)
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0] == (fw_file1, False, 3)
        assert patch_fw_callback.call_args_list[1][0] == (fw_file2, False, 4)

    with pytest.raises(IMException) as exc_info:
        mc.communication.load_firmware_ecat("", TEST_ENSEMBLE_FW_FILE, slave=5)
    assert str(exc_info.value) == "The selected drive is not part of the ensemble."


@pytest.mark.virtual
def test_load_ensemble_fw_canopen(mocker):
    class MockDictionary:
        def __init__(self) -> None:
            self.path = "path_to_dictionary"

    class MockCanopenServo(CanopenServo):
        def __init__(self, node_id) -> None:
            self.target = node_id
            self._dictionary = MockDictionary()

    servos = {}
    for node_id in range(1, 6):
        servos[str(node_id)] = MockCanopenServo(node_id)

    mc = MotionController()
    net = CanopenNetwork(CAN_DEVICE.KVASER)
    net.servos = list(servos.values())
    mocker.patch("ingeniamotion.motion_controller.MotionController._get_network", return_value=net)
    mocker.patch("ingenialink.canopen.network.CanopenNetwork.connect_to_slave")
    mocker.patch("ingenialink.canopen.network.CanopenNetwork.disconnect_from_slave")
    mc._get_drive = lambda x: servos[x]
    mc.servos = servos

    product_code = 123456
    slaves_info = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(product_code, 16781876),
            3: SlaveInfo(product_code, 4660),
            4: SlaveInfo(product_code, 16781876),
            5: SlaveInfo(654321, 1236),
        }
    )
    temp_path = os.path.abspath("ensemble_temp")
    fw_file1 = os.path.join(temp_path, "cap-net-1-e_2.4.0.lfu")
    fw_file2 = os.path.join(temp_path, "cap-net-2-e_2.4.0.lfu")
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.scan_slaves_info", return_value=slaves_info
    )
    for slave in [1, 2]:
        patch_fw_callback = mocker.patch("ingenialink.canopen.network.CanopenNetwork.load_firmware")
        mc.communication.load_firmware_canopen(TEST_ENSEMBLE_FW_FILE, servo=str(slave))
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0][:2] == (1, fw_file1)
        assert patch_fw_callback.call_args_list[1][0][:2] == (2, fw_file2)

    for slave in [3, 4]:
        patch_fw_callback = mocker.patch("ingenialink.canopen.network.CanopenNetwork.load_firmware")
        mc.communication.load_firmware_canopen(TEST_ENSEMBLE_FW_FILE, servo=str(slave))
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0][:2] == (3, fw_file1)
        assert patch_fw_callback.call_args_list[1][0][:2] == (4, fw_file2)

    with pytest.raises(IMException) as exc_info:
        mc.communication.load_firmware_canopen(TEST_ENSEMBLE_FW_FILE, servo="5")
    assert str(exc_info.value) == "The selected drive is not part of the ensemble."



from typing import TYPE_CHECKING
import ingenialogger

from ingeniamotion.enums import OperationMode, HomingMode

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController
from ingeniamotion.metaclass import MCMetaClass, DEFAULT_AXIS, DEFAULT_SERVO


class Homing(metaclass=MCMetaClass):

    HOMING_MODE_REGISTER = "HOM_MODE"
    HOMING_OFFSET_REGISTER = "HOM_OFFSET"
    HOMING_TIMEOUT_REGISTER = "HOM_SEQ_TIMEOUT"
    POSITIVE_HOMING_SWITCH_REGISTER = "IO_IN_POS_HOM_SWITCH"
    NEGATIVE_HOMING_SWITCH_REGISTER = "IO_IN_NEG_HOM_SWITCH"
    HOMING_SEARCH_VELOCITY_REGISTER = "HOM_SPEED_SEARCH"
    HOMING_ZERO_VELOCITY_REGISTER = "HOM_SPEED_ZERO"
    HOMING_INDEX_PULSE_SOURCE_REGISTER = "HOM_IDX_PULSE_SRC"

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)

    def set_homing_mode(
        self, homing_mode: HomingMode, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set homing mode.

        Args:
            homing_mode : homing mode.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        self.mc.communication.set_register(self.HOMING_MODE_REGISTER, homing_mode, servo, axis)

    def set_homing_offset(
        self, homing_offset: int, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set homing offset configuration.

        Args:
            homing_offset : homing offset.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        self.mc.communication.set_register(self.HOMING_OFFSET_REGISTER, homing_offset, servo, axis)

    def set_homing_timeout(
        self, timeout_ms: int, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set homing timeout configuration.

        Args:
            timeout_ms : homing timeout in milliseconds.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        self.mc.communication.set_register(self.HOMING_TIMEOUT_REGISTER, timeout_ms, servo, axis)

    def __check_motor_phasing(self, servo: str, axis: int) -> None:
        if not self.mc.configuration.is_commutation_feedback_aligned(servo=servo, axis=axis):
            self.logger.warning(
                "Motor must be well phased before run any homing.",
                axis=axis,
                drive=self.mc.servo_name(servo),
            )

    def homing_on_current_position(
        self, hom_offset: int, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Do current position homing.

        Args:
            hom_offset : homing offset.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        # Save previous mode
        prev_op_mode = self.mc.motion.get_operation_mode(servo, axis)

        self.mc.communication.set_register(
            self.HOMING_MODE_REGISTER, HomingMode.CURRENT_POSITION, servo, axis
        )
        self.mc.communication.set_register(self.HOMING_OFFSET_REGISTER, hom_offset, servo, axis)
        self.mc.motion.set_operation_mode(OperationMode.HOMING, servo, axis)

        # Perform the homing
        self.mc.motion.target_latch(servo, axis)
        # Restore op mode
        self.mc.motion.set_operation_mode(prev_op_mode, servo, axis)

    def homing_on_switch_limit(
        self,
        hom_offset: int,
        direction: int,
        switch: int,
        timeout_ms: int,
        lim_vel: int,
        zero_vel: int,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        motor_enable: bool = True,
    ) -> None:
        """Do homing on switch limit.

        .. note::
            Motor must be well phased before run any homing.

        Args:
            hom_offset : homing offset.
            direction : direction. ``1`` is positive, ``0`` is negative.
            switch : switch index.
            timeout_ms : homing timeout in milliseconds.
            lim_vel : speed to search for the switch, in mrev/s.
            zero_vel : speed to search for the actual homing point, in mrev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            motor_enable : if ``True`` do motor enable. ``True`` by default.

        """
        self.mc.motion.set_operation_mode(OperationMode.HOMING, servo, axis)
        if direction > 0:
            # Positive direction
            self.set_homing_mode(HomingMode.POSITIVE_LIMIT_SWITCH, servo, axis)
            self.mc.communication.set_register(
                self.POSITIVE_HOMING_SWITCH_REGISTER, switch, servo, axis
            )
        else:
            # Negative direction
            self.set_homing_mode(HomingMode.NEGATIVE_LIMIT_SWITCH, servo, axis)
            self.mc.communication.set_register(
                self.NEGATIVE_HOMING_SWITCH_REGISTER, switch, servo, axis
            )
        self.set_homing_offset(hom_offset, servo, axis)
        self.set_homing_timeout(timeout_ms, servo, axis)
        self.mc.communication.set_register(
            self.HOMING_SEARCH_VELOCITY_REGISTER, lim_vel, servo, axis
        )
        self.mc.communication.set_register(
            self.HOMING_ZERO_VELOCITY_REGISTER, zero_vel, servo, axis
        )
        self.__check_motor_phasing(servo, axis)
        # Perform the homing
        if motor_enable:
            self.mc.motion.motor_enable(servo, axis)
            self.mc.motion.target_latch(servo, axis)

    def homing_on_index_pulse(
        self,
        hom_offset: int,
        direction: int,
        index: int,
        timeout_ms: int,
        zero_vel: int,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        motor_enable: bool = True,
    ) -> None:
        """Do homing on index pulse.

        .. note::
            Motor must be well phased before run any homing.

        Args:
            hom_offset : homing offset.
            direction : direction. ``1`` is positive, ``0`` is negative.
            index : select incremental encoder, ``0`` for incremental encoder 1,
             ``1`` for incremental encoder 2.
            timeout_ms : homing timeout in milliseconds.
            zero_vel : speed to search for the actual homing point, in mrev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            motor_enable : if ``True`` do motor enable. ``True`` by default.
        """
        self.mc.motion.set_operation_mode(OperationMode.HOMING, servo, axis)
        if direction > 0:
            # Positive direction
            self.set_homing_mode(HomingMode.POSITIVE_IDX_PULSE, servo, axis)
        else:
            # Negative direction
            self.set_homing_mode(HomingMode.NEGATIVE_IDX_PULSE, servo, axis)
        self.set_homing_offset(hom_offset, servo, axis)
        self.set_homing_timeout(timeout_ms, servo, axis)
        self.mc.communication.set_register(
            self.HOMING_INDEX_PULSE_SOURCE_REGISTER, index, servo, axis
        )
        self.mc.communication.set_register(
            self.HOMING_ZERO_VELOCITY_REGISTER, zero_vel, servo, axis
        )
        self.__check_motor_phasing(servo, axis)
        # Perform the homing
        if motor_enable:
            self.mc.motion.motor_enable(servo, axis)
        self.mc.motion.target_latch(servo, axis)

    def homing_on_switch_limit_and_index_pulse(
        self,
        hom_offset: int,
        direction: int,
        switch: int,
        index: int,
        timeout_ms: int,
        lim_vel: int,
        zero_vel: int,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        motor_enable: bool = True,
    ) -> None:
        """
        Do homing on switch limit and index pulse.

        .. note::
            Motor must be well phased before run any homing.

        Args:
            hom_offset : homing offset.
            direction : direction. ``1`` is positive, ``0`` is negative.
            switch : switch index.
            index : select incremental encoder, ``0`` for incremental encoder 1,
             ``1`` for incremental encoder 2.
            timeout_ms : homing timeout in milliseconds.
            lim_vel : speed to search for the switch, in mrev/s.
            zero_vel : speed to search for the actual homing point, in mrev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            motor_enable : if ``True`` do motor enable. ``True`` by default.
        """
        self.mc.motion.set_operation_mode(OperationMode.HOMING, servo, axis)
        if direction > 0:
            # Positive direction
            self.set_homing_mode(HomingMode.POSITIVE_LIMIT_SWITCH_IDX_PULSE, servo, axis)
            self.mc.communication.set_register(
                self.POSITIVE_HOMING_SWITCH_REGISTER, switch, servo, axis
            )
        else:
            # Negative direction
            self.set_homing_mode(HomingMode.NEGATIVE_LIMIT_SWITCH_IDX_PULSE, servo, axis)
            self.mc.communication.set_register(
                self.NEGATIVE_HOMING_SWITCH_REGISTER, switch, servo, axis
            )
        self.set_homing_offset(hom_offset, servo, axis)
        self.set_homing_timeout(timeout_ms, servo, axis)
        self.mc.communication.set_register(
            self.HOMING_INDEX_PULSE_SOURCE_REGISTER, index, servo, axis
        )
        self.mc.communication.set_register(
            self.HOMING_SEARCH_VELOCITY_REGISTER, lim_vel, servo, axis
        )
        self.mc.communication.set_register(
            self.HOMING_ZERO_VELOCITY_REGISTER, zero_vel, servo, axis
        )
        self.__check_motor_phasing(servo, axis)
        # Perform the homing
        if motor_enable:
            self.mc.motion.motor_enable(servo, axis)
        self.mc.motion.target_latch(servo, axis)



from typing import TYPE_CHECKING, Dict, Optional, Union
import ingenialogger

from ingeniamotion.enums import SensorType, SeverityLevel

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController
from ingeniamotion.wizard_tests.feedbacks_tests.feedback_test import Feedbacks
from ingeniamotion.wizard_tests.feedbacks_tests.absolute_encoder1_test import AbsoluteEncoder1Test
from ingeniamotion.wizard_tests.feedbacks_tests.digital_incremental1_test import (
    DigitalIncremental1Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.digital_hall_test import DigitalHallTest
from ingeniamotion.wizard_tests.feedbacks_tests.secondary_ssi_test import SecondarySSITest
from ingeniamotion.wizard_tests.feedbacks_tests.absolute_encoder2_test import AbsoluteEncoder2Test
from ingeniamotion.wizard_tests.feedbacks_tests.digital_incremental2_test import (
    DigitalIncremental2Test,
)
from ingeniamotion.wizard_tests.phase_calibration import Phasing
from ingeniamotion.wizard_tests.phasing_check import PhasingCheck
from ingeniamotion.wizard_tests.sto import STOTest
from ingeniamotion.wizard_tests.brake import Brake
from ingeniamotion.metaclass import MCMetaClass, DEFAULT_AXIS, DEFAULT_SERVO
from ingeniamotion.wizard_tests.feedbacks_tests.dc_feedback_polarity_test import (
    DCFeedbacksPolarityTest,
)
from ingeniamotion.wizard_tests.feedbacks_tests.dc_feedback_resolution_test import (
    DCFeedbacksResolutionTest,
)


class DriveTests(metaclass=MCMetaClass):
    __sensors = {
        SensorType.ABS1: AbsoluteEncoder1Test,
        SensorType.QEI: DigitalIncremental1Test,
        SensorType.HALLS: DigitalHallTest,
        SensorType.SSI2: SecondarySSITest,
        SensorType.BISSC2: AbsoluteEncoder2Test,
        SensorType.QEI2: DigitalIncremental2Test,
    }

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)

    def digital_halls_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes the digital halls feedback test given a target servo and
        axis. By default test will make changes in some drive registers like
        feedback polarity and others suggested registers. To avoid it, set
        ``apply_changes`` to ``False``.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.
            apply_changes : if ``True``, test applies changes to the
                servo, if ``False`` it does not. ``True`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers":
                        {"FBK_DIGHALL_POLARITY": 0},
                    # (str) Human readable result message
                    "result_message": "Feedback test pass successfully"
                }

        Raises:
            TestError: In case the servo or setup configuration makes
                impossible fulfilling the test
        """
        return self.__feedback_test(SensorType.HALLS, servo, axis, apply_changes)

    def incremental_encoder_1_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes the incremental encoder 1 feedback test given a target servo
        and axis. By default test will make changes in some drive registers
        like feedback polarity and other suggested registers. To avoid it, set
        ``apply_changes`` to ``False``.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.
            apply_changes : if ``True``, test applies changes to the
                servo, if ``False`` it does not. ``True`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers":
                        {"FBK_DIGENC1_POLARITY": 0},
                    # (str) Human readable result message
                    "result_message": "Feedback test pass successfully"
                }

        Raises:
            TestError: In case the servo or setup configuration makes
                impossible fulfilling the test
        """
        return self.__feedback_test(SensorType.QEI, servo, axis, apply_changes)

    def incremental_encoder_2_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes incremental encoder 2 feedback test given a target servo
        and axis. By default test will make changes in some drive registers
        like feedback polarity and other suggested registers. To avoid it,
        set ``apply_changes`` to ``False``.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.
            apply_changes : if ``True``, test applies changes to the
                servo, if ``False`` it does not. ``True`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers":
                        {"FBK_DIGENC2_POLARITY": 0},
                    # (str) Human readable result message
                    "result_message": "Feedback test pass successfully"
                }

        Raises:
            TestError: In case the servo or setup configuration makes
                impossible fulfilling the test
        """
        return self.__feedback_test(SensorType.QEI2, servo, axis, apply_changes)

    def absolute_encoder_1_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes absolute encoder 1 feedback test given a target servo and axis.
        To know more about it see :func:`digital_halls_test`.
        """
        return self.__feedback_test(SensorType.ABS1, servo, axis, apply_changes)

    def absolute_encoder_2_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes absolute encoder 2 feedback test given a target servo and axis.
        To know more about it see :func:`digital_halls_test`.
        """
        return self.__feedback_test(SensorType.BISSC2, servo, axis, apply_changes)

    def secondary_ssi_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes secondary SSI feedback test given a target servo and axis.
        To know more about it see :func:`digital_halls_test`.
        """
        return self.__feedback_test(SensorType.SSI2, servo, axis, apply_changes)

    def get_feedback_test(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Feedbacks:
        return self.__sensors[feedback](self.mc, servo, axis)

    def __feedback_test(
        self,
        feedback: SensorType,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        apply_changes: bool = True,
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        output = self.get_feedback_test(feedback, servo, axis).run()
        if (
            apply_changes
            and output is not None
            and output["result_severity"] == SeverityLevel.SUCCESS
        ):
            if not isinstance(output["suggested_registers"], Dict):
                raise TypeError("Suggested registers has to be a dictionary")
            for key, value in output["suggested_registers"].items():
                self.mc.communication.set_register(key, value, servo=servo, axis=axis)
            self.logger.debug(
                "Feedback test changes applied", axis=axis, drive=self.mc.servo_name(servo)
            )
        return output

    def commutation(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, apply_changes: bool = True
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes a commutation calibration given a target servo and axis.
        By default commutation will make changes in some drive registers
        like commutation angle offset and other suggested registers.
        To avoid it, set ``apply_changes`` to ``False``.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.
            apply_changes : if ``True``, test applies changes to the
                servo, if ``False`` it does not. ``True`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers":
                        {"COMMU_ANGLE_OFFSET": 0.12},
                    # (str) Human readable result message
                    "result_message": "Phasing process finished successfully"
                }

        Raises:
            TestError: If servo or setup configuration makes impossible
                complete the calibration.
            TypeError: If some parameter has a wrong type.
        """
        commutation = Phasing(self.mc, servo, axis)
        output = commutation.run()
        if (
            apply_changes
            and output is not None
            and output["result_severity"] == SeverityLevel.SUCCESS
        ):
            if not isinstance(output["suggested_registers"], Dict):
                raise TypeError("Suggested registers have to be a dictionary")
            for key, value in output["suggested_registers"].items():
                self.mc.communication.set_register(key, value, servo=servo, axis=axis)
            self.logger.debug(
                "Commutation changes applied", axis=axis, drive=self.mc.servo_name(servo)
            )
        return output

    def phasing_check(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """
        Checks servo phasing.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers": {},
                    # (str) Human readable result message
                    "result_message": "Phasing process finished successfully"
                }
        """
        phasing_check = PhasingCheck(self.mc, servo, axis)
        return phasing_check.run()

    def sto_test(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """
        Check STO

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers": {},
                    # (str) Human readable result message
                    "result_message": "Phasing process finished successfully"
                }
        """
        sto_test = STOTest(self.mc, servo, axis)
        return sto_test.run()

    def brake_test(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> Brake:
        """
        Run brake test.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Instance of Brake test. Call ``Brake.finish()`` to end the test.
        """
        brake_test = Brake(self.mc, servo, axis)
        brake_test.run()
        return brake_test

    def polarity_feedback_single_phase_test(
        self,
        feedback: SensorType,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        apply_changes: bool = True,
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes polarity feedback test for single phase motors given a target servo
        and axis. By default, test will make changes in feedback polarity. To avoid it,
        set ``apply_changes`` to ``False``.

        Args:
            feedback: feedback sensor type
            servo: servo alias to reference it. ``default`` by default.
            axis: axis that will run the test. ``1`` by default.
            apply_changes: if ``True``, test applies changes to the
                servo, if ``False`` it does not. ``True`` by default.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (dict) Suggested register values
                    "suggested_registers":
                        {"FBK_DIGENC2_POLARITY": 0},
                    # (str) Human readable result message
                    "result_message": "Feedback test pass successfully"
                }

        Raises:
            TestError: In case the servo or setup configuration makes
                impossible fulfilling the test
            TypeError: If some parameter has a wrong type.
        """
        dc_feedback_polarity_test = DCFeedbacksPolarityTest(self.mc, feedback, servo, axis)
        output = dc_feedback_polarity_test.run()
        if (
            apply_changes
            and output is not None
            and output["result_severity"] == SeverityLevel.SUCCESS
        ):
            if not isinstance(output["suggested_registers"], Dict):
                raise TypeError("Suggested registers have to be a dictionary")
            for key, value in output["suggested_registers"].items():
                self.mc.communication.set_register(key, value, servo=servo, axis=axis)
            self.logger.debug(
                "Single phase feedback polarity test changes applied",
                axis=axis,
                drive=self.mc.servo_name(servo),
            )
        return output

    def resolution_feedback_single_phase_test(
        self,
        feedback: SensorType,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        kp: Optional[float] = None,
        ki: Optional[float] = None,
        kd: Optional[float] = None,
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        """Executes resolution feedback test for single phase motors given a target servo
        and axis. This test needs a human check to ensure the feedback is well configured.
        The test will move the motor with the number of counts set in the feedback resolution,
        if the motor does not move exactly one revolution this means that
        the feedback is not configured correctly.

        Args:
            feedback: feedback sensor type
            servo: servo alias to reference it. ``default`` by default.
            axis: axis that will run the test. ``1`` by default.
            kp: overrides test velocity Kp. If ``None`` use test default Kp value.
                At the end of the test initial drive value is restored.
            ki: if ki is ``None`` is ignored, overrides test velocity Ki.
                If ``None`` use test default Ki value.
                At the end of the test initial drive value is restored.
            kd: if kd is ``None`` is ignored, overrides test velocity Kd.
                If ``None`` use test default Kd value.
                At the end of the test initial drive value is restored.

        Returns:
            Dictionary with the result of the test::

                {
                    # (int) Result code
                    "result_severity": 0,
                    # (str) Human readable result message
                    "result_message": "Feedback test pass successfully"
                }

        Raises:
            TestError: In case the servo or setup configuration makes
                impossible fulfilling the test
        """
        dc_feedback_resolution_test = DCFeedbacksResolutionTest(
            self.mc, feedback, servo, axis, kp=kp, ki=ki, kd=kd
        )
        return dc_feedback_resolution_test.run()



from enum import IntEnum, Enum
from ingenialink.enums.register import REG_DTYPE, REG_ACCESS
from ingenialink.canopen.network import CAN_BAUDRATE, CAN_DEVICE


class OperationMode(IntEnum):
    """Operation Mode Enum"""

    VOLTAGE = 0x00
    CURRENT_AMPLIFIER = 0x01
    CURRENT = 0x02
    CYCLIC_CURRENT = 0x22
    VELOCITY = 0x03
    PROFILE_VELOCITY = 0x13
    CYCLIC_VELOCITY = 0x23
    POSITION = 0x04
    PROFILE_POSITION = 0x14
    CYCLIC_POSITION = 0x24
    PROFILE_POSITION_S_CURVE = 0x44
    INTERPOLATED_POSITION = 0xA4
    PVT = 0xB4
    HOMING = 0x113
    TORQUE = 0x05
    CYCLIC_TORQUE = 0x25


class Protocol(IntEnum):
    """Communication protocol"""

    TCP = 1
    UDP = 2


class HomingMode(IntEnum):
    """Homing modes"""

    CURRENT_POSITION = 0
    POSITIVE_LIMIT_SWITCH = 1
    NEGATIVE_LIMIT_SWITCH = 2
    POSITIVE_IDX_PULSE = 3
    NEGATIVE_IDX_PULSE = 4
    POSITIVE_LIMIT_SWITCH_IDX_PULSE = 5
    NEGATIVE_LIMIT_SWITCH_IDX_PULSE = 6


class MonitoringSoCType(IntEnum):
    """Monitoring start of condition type"""

    TRIGGER_EVENT_AUTO = 0
    """No trigger"""
    TRIGGER_EVENT_FORCED = 1
    """Forced trigger"""
    TRIGGER_EVENT_EDGE = 2
    """Edge trigger"""


class MonitoringSoCConfig(IntEnum):
    TRIGGER_CONFIG_RISING_OR_FALLING = 0
    """Rising or falling edge trigger"""
    TRIGGER_CONFIG_RISING = 1
    """Rising edge trigger"""
    TRIGGER_CONFIG_FALLING = 2
    """Falling edge trigger"""


class MonitoringProcessStage(IntEnum):
    """Monitoring process stage"""

    INIT_STAGE = 0x0
    """Init stage"""
    FILLING_DELAY_DATA = 0x2
    """Filling delay data"""
    WAITING_FOR_TRIGGER = 0x4
    """Waiting for trigger"""
    DATA_ACQUISITION = 0x6
    """Data acquisition"""
    END_STAGE = 0x8
    """End stage"""


class SensorType(IntEnum):
    """Summit series feedback type enum"""

    ABS1 = 1
    """Absolute encoder 1"""
    INTGEN = 3
    """Internal generator"""
    QEI = 4
    """Digital/Incremental encoder 1"""
    HALLS = 5
    """Digital halls"""
    SSI2 = 6
    """Secondary SSI"""
    BISSC2 = 7
    """Absolute encoder 2"""
    QEI2 = 8
    """Digital/Incremental encoder 2"""


class SensorCategory(IntEnum):
    """Feedback category enum"""

    ABSOLUTE = 0
    INCREMENTAL = 1


class PhasingMode(IntEnum):
    """Phasing modes"""

    NON_FORCED = 0
    """Non forced"""
    FORCED = 1
    """Forced"""
    NO_PHASING = 2
    """No phasing"""


class GeneratorMode(IntEnum):
    """Generator modes"""

    CONSTANT = 0
    """Constant"""
    SAW_TOOTH = 1
    """Saw tooth"""
    SQUARE = 2
    """Square"""


class MonitoringVersion(IntEnum):
    """Monitoring version"""

    MONITORING_V1 = 0
    """Monitoring V1 used for Everest 1.8.1 and older."""
    MONITORING_V2 = 1
    """Monitoring V2 used for Capitan and some custom low-power drivers."""
    MONITORING_V3 = 2
    """Monitoring V3 used for Everest and Capitan newer than 1.8.1."""


class SeverityLevel(IntEnum):
    """Test result enum"""

    SUCCESS = 0
    WARNING = 1
    FAIL = 2


class COMMUNICATION_TYPE(IntEnum):
    Canopen = 0
    Ethernet = 1
    Ethercat = 2


class FeedbackPolarity(IntEnum):
    """Feedback polarity enum"""

    NORMAL = 0
    REVERSED = 1


class CommutationMode(IntEnum):
    """Commutation Mode Enum"""

    SINUSOIDAL = 0
    TRAPEZOIDAL = 1
    SINGLE_PHASE = 2


class FilterType(IntEnum):
    """
    Biquad filter type.
    """

    DISABLED = 0
    """ Filter disabled """
    LOWPASS = 1
    """ Low-pass filter """
    HIGHPASS = 2
    """ High-pass filter """
    BANDPASS = 3
    """ Band-pass filter """
    PEAK = 4
    """ Peak filter """
    NOTCH = 5
    """ Notch filter """
    LOWSHELF = 6
    """ Low Shelf filter """
    HIGHSHELF = 7
    """ High Shelf filter """


class FilterSignal(Enum):
    """Signal to configure filter."""

    POSITION_FEEDBACK = "POS_FBK"
    """Position feedback."""
    POSITION_REFERENCE = "POS_REF"
    """Position reference."""
    VELOCITY_FEEDBACK = "VEL_FBK"
    """Velocity feedback."""
    VELOCITY_REFERENCE = "VEL_REF"
    """Velocity reference."""
    CURRENT_FEEDBACK = "CUR_FBK"
    """Current feedback."""
    CURRENT_REFERENCE = "CUR_REF"
    """Current reference."""


class FilterNumber(IntEnum):
    """Filter number (1 or 2)."""

    FILTER1 = 1
    FILTER2 = 2


enums = list(globals().keys())
enums.remove("IntEnum")
__all__ = enums



from enum import IntEnum
import time
from typing import TYPE_CHECKING

import ingenialogger

from ingeniamotion.wizard_tests.base_test import BaseTest, TestError
from ingeniamotion.metaclass import DEFAULT_SERVO, DEFAULT_AXIS
from ingeniamotion.enums import SensorType, OperationMode, PhasingMode, SeverityLevel

if TYPE_CHECKING:
    from ingeniamotion import MotionController


class PhasingCheck(BaseTest):

    MAX_ALLOWED_ANGLE_MOVE = 15
    INITIAL_ANGLE = 180.0
    INITIAL_ANGLE_HALLS = 240.0
    CURRENT_SLOPE = 0.4

    PHASING_ACCURACY_REGISTER = "COMMU_PHASING_ACCURACY"
    PHASING_TIMEOUT_REGISTER = "COMMU_PHASING_TIMEOUT"
    MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER = "COMMU_PHASING_MAX_CURRENT"

    BACKUP_REGISTERS = ["DRV_OP_CMD", "CL_CUR_Q_SET_POINT", "CL_CUR_D_SET_POINT"]

    class ResultType(IntEnum):
        SUCCESS = 0
        WRONG_PHASING = -1

    result_description = {
        ResultType.SUCCESS: "Motor is well phased.",
        ResultType.WRONG_PHASING: "Phasing check failed. Wrong motor phasing.",
    }

    def __init__(
        self, mc: "MotionController", servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.axis = axis
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))
        self.backup_registers_names = self.BACKUP_REGISTERS

    @BaseTest.stoppable
    def setup(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.mc.motion.set_operation_mode(OperationMode.CURRENT, servo=self.servo, axis=self.axis)
        self.mc.motion.set_current_quadrature(0, servo=self.servo, axis=self.axis)
        self.mc.motion.set_current_direct(0, servo=self.servo, axis=self.axis)

    @BaseTest.stoppable
    def vary_current_and_check_position(self, current_sign: int, max_current: int) -> ResultType:
        init_pos = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
        pos_resolution = self.mc.configuration.get_position_feedback_resolution(
            servo=self.servo, axis=self.axis
        )
        is_ok = True
        total_time = max_current / self.CURRENT_SLOPE
        for value in self.mc.motion.ramp_generator(
            0, current_sign * max_current, total_time, interval=0.1
        ):
            self.mc.motion.set_current_direct(value, servo=self.servo, axis=self.axis)
            is_ok = self.check_position(init_pos, pos_resolution)
            if not is_ok:
                break
        self.mc.motion.set_current_direct(0, servo=self.servo, axis=self.axis)
        if not is_ok:
            return self.ResultType.WRONG_PHASING
        return self.ResultType.SUCCESS

    @BaseTest.stoppable
    def check_position(self, init_pos: int, resolution: int) -> bool:
        current_pos = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
        allowed_move = resolution * self.MAX_ALLOWED_ANGLE_MOVE / 360
        result_check = float(abs(current_pos - init_pos)) < allowed_move
        return result_check

    @BaseTest.stoppable
    def analyse_phasing_bit_and_force_to_high(self) -> None:
        if not self.mc.configuration.is_commutation_feedback_aligned(
            servo=self.servo, axis=self.axis
        ):
            phasing_mode = self.mc.configuration.get_phasing_mode(servo=self.servo, axis=self.axis)
            if phasing_mode == PhasingMode.FORCED:
                self.forced_phasing()
            else:
                self.not_forced_phasing()

    @BaseTest.stoppable
    def define_phasing_steps(self) -> int:
        pha_accuracy = self.mc.communication.get_register(
            self.PHASING_ACCURACY_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(pha_accuracy, int):
            raise TypeError(f"{self.PHASING_ACCURACY_REGISTER} has to be an integer")
        delta = 3 * pha_accuracy / 1000
        ref_feedback = self.mc.configuration.get_reference_feedback(
            servo=self.servo, axis=self.axis
        )

        num_of_steps = 1
        if ref_feedback == SensorType.HALLS:
            actual_angle = self.INITIAL_ANGLE_HALLS
        else:
            actual_angle = self.INITIAL_ANGLE
        while actual_angle > delta:
            actual_angle = actual_angle / 2
            num_of_steps += 1
        return num_of_steps

    @BaseTest.stoppable
    def forced_phasing(self) -> None:
        num_of_steps = self.define_phasing_steps()
        phasing_bit_latched = False
        phasing_timeout = self.mc.communication.get_register(
            self.PHASING_TIMEOUT_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(phasing_timeout, int):
            raise TypeError(f"{self.PHASING_TIMEOUT_REGISTER} has to be an integer")
        timeout = time.time() + (num_of_steps * phasing_timeout / 1000)
        while time.time() < timeout:
            self.stoppable_sleep(0.1)
            if self.mc.configuration.is_commutation_feedback_aligned(
                servo=self.servo, axis=self.axis
            ):
                phasing_bit_latched = True
                break
        if not phasing_bit_latched:
            raise TestError("Motor phasing fail")

    @BaseTest.stoppable
    def not_forced_phasing(self) -> None:
        self.logger.info(
            "Increasing slowly current quadrature set-point until phasing bit is latched"
        )
        phasing_bit_latched = False
        i = 0.0
        delta_i = 0.1
        while not phasing_bit_latched:
            if not self.mc.configuration.is_motor_enabled(servo=self.servo, axis=self.axis):
                self.mc.errors.get_last_buffer_error(servo=self.servo, axis=self.axis)
                self.show_error_message()
            self.mc.motion.set_current_quadrature(i, servo=self.servo, axis=self.axis)
            self.stoppable_sleep(0.1)
            i += delta_i
            if self.mc.configuration.is_commutation_feedback_aligned(
                servo=self.servo, axis=self.axis
            ):
                phasing_bit_latched = True
        self.mc.motion.set_current_quadrature(0, servo=self.servo, axis=self.axis)
        self.logger.info("Wait until motor has stopped")
        self.mc.motion.wait_for_velocity(0, servo=self.servo, axis=self.axis)

    @BaseTest.stoppable
    def check_motor_commutation(self) -> ResultType:
        phasing_current = self.mc.communication.get_register(
            self.MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(phasing_current, float):
            raise TypeError(f"{self.MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER} has to be an integer")
        max_test_current = round(phasing_current, 2)
        ref_feedback = self.mc.configuration.get_reference_feedback(
            servo=self.servo, axis=self.axis
        )
        comm_feedback = self.mc.configuration.get_commutation_feedback(
            servo=self.servo, axis=self.axis
        )

        # With Halls only, this test makes no sense
        if ref_feedback == SensorType.HALLS and comm_feedback == SensorType.HALLS:
            return self.ResultType.SUCCESS

        self.logger.info(
            "{} {} {}".format(
                "Slowly increasing Current Direct Set-point until", max_test_current, "A"
            )
        )
        result = self.vary_current_and_check_position(1, max_test_current)
        if result != self.ResultType.SUCCESS:
            return result
        self.logger.info(
            "{} {} {}".format(
                "Slowly decreasing Current Direct Set-point until", -max_test_current, "A"
            )
        )
        return self.vary_current_and_check_position(-1, max_test_current)

    @BaseTest.stoppable
    def loop(self) -> ResultType:
        self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)
        self.analyse_phasing_bit_and_force_to_high()
        return self.check_motor_commutation()

    def teardown(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)

    def get_result_msg(self, output: ResultType) -> str:
        return self.result_description[output]

    def get_result_severity(self, output: ResultType) -> SeverityLevel:
        if output < self.ResultType.SUCCESS:
            return SeverityLevel.FAIL
        else:
            return SeverityLevel.SUCCESS



from enum import IntEnum
import time
from typing import TYPE_CHECKING, Optional
import ingenialogger

from ingeniamotion.enums import (
    PhasingMode,
    OperationMode,
    SensorCategory,
    SensorType,
    SeverityLevel,
)
from ingeniamotion.exceptions import IMRegisterNotExist
from ingeniamotion.wizard_tests.base_test import BaseTest, TestError
from ingeniamotion.enums import CommutationMode

if TYPE_CHECKING:
    from ingeniamotion import MotionController


class Phasing(BaseTest):
    INTERNAL_GENERATOR_VALUE = 3
    INITIAL_ANGLE = 180.0
    INITIAL_ANGLE_HALLS = 240.0
    PHASING_CURRENT_PERCENTAGE = 0.4
    PHASING_CURRENT_PERCENTAGE_GEAR = 0.8
    PHASING_TIMEOUT_DEFAULT = 2000
    PHASING_ACCURACY_HALLS_DEFAULT = 60000
    PHASING_ACCURACY_DEFAULT = 3600

    MAX_CURRENT_REGISTER = "CL_CUR_REF_MAX"
    RATED_CURRENT_REGISTER = "MOT_RATED_CURRENT"
    PHASING_ACCURACY_REGISTER = "COMMU_PHASING_ACCURACY"
    PHASING_TIMEOUT_REGISTER = "COMMU_PHASING_TIMEOUT"
    MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER = "COMMU_PHASING_MAX_CURRENT"
    COMMUTATION_ANGLE_OFFSET_REGISTER = "COMMU_ANGLE_OFFSET"

    BACKUP_REGISTERS = [
        "CL_POS_FBK_SENSOR",
        "DRV_OP_CMD",
        "CL_CUR_Q_SET_POINT",
        "CL_CUR_D_SET_POINT",
        "FBK_GEN_MODE",
        "FBK_GEN_FREQ",
        "FBK_GEN_GAIN",
        "FBK_GEN_OFFSET",
        "COMMU_ANGLE_SENSOR",
        "FBK_GEN_CYCLES",
        "COMMU_PHASING_MAX_CURRENT",
        "COMMU_PHASING_TIMEOUT",
        "COMMU_PHASING_ACCURACY",
        "COMMU_PHASING_MODE",
        "MOT_COMMU_MOD",
        "COMMU_ANGLE_INTEGRITY1_OPTION",
        "COMMU_ANGLE_INTEGRITY2_OPTION",
    ]

    class ResultType(IntEnum):
        SUCCESS = 0
        FAIL = -1

    result_description = {
        ResultType.SUCCESS: "Phasing process finished successfully",
        ResultType.FAIL: "Phasing process has failed.  Review your motor configuration.",
    }

    def __init__(
        self,
        mc: "MotionController",
        servo: str,
        axis: int,
        default_current: bool = True,
        default_timeout: bool = True,
        default_accuracy: bool = True,
    ) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.axis = axis
        self.backup_registers_names = self.BACKUP_REGISTERS
        self.comm: Optional[SensorType] = None
        self.ref: Optional[SensorType] = None
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))

        self.default_phasing_current = default_current
        self.default_phasing_timeout = default_timeout
        self.default_phasing_accuracy = default_accuracy

        self.pha_current: float = 0.0
        self.pha_timeout: float = 0.0
        self.pha_accuracy: float = 0.0

    @BaseTest.stoppable
    def check_input_data(self) -> None:

        max_current_drive = self.mc.communication.get_register(
            self.MAX_CURRENT_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(max_current_drive, float):
            raise TypeError("Max. current of the drive has to be a float")
        max_current_motor = self.mc.communication.get_register(
            self.RATED_CURRENT_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(max_current_motor, float):
            raise TypeError("Rated current has to be a float")
        max_test_current = min(max_current_drive, max_current_motor)

        if self.default_phasing_current:
            pos_vel_ratio = self.mc.configuration.get_pos_to_vel_ratio(
                servo=self.servo, axis=self.axis
            )
            if pos_vel_ratio == 1:
                self.pha_current = self.PHASING_CURRENT_PERCENTAGE * max_test_current
            else:
                self.pha_current = self.PHASING_CURRENT_PERCENTAGE_GEAR * max_test_current
            self.mc.communication.set_register(
                self.MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER,
                value=self.pha_current,
                servo=self.servo,
                axis=self.axis,
            )
        else:
            pha_current = self.mc.communication.get_register(
                self.MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER, servo=self.servo, axis=self.axis
            )
            if not isinstance(pha_current, float):
                raise TypeError(
                    f"{self.MAX_CURRENT_ON_PHASING_SEQUENCE_REGISTER} has to be a float"
                )
            self.pha_current = pha_current
            if self.pha_current > max_test_current:
                raise TestError("Defined phasing current is higher than configured maximum current")
        if self.default_phasing_timeout:
            self.mc.communication.set_register(
                self.PHASING_TIMEOUT_REGISTER,
                self.PHASING_TIMEOUT_DEFAULT,
                servo=self.servo,
                axis=self.axis,
            )
        if self.default_phasing_accuracy:
            if SensorType.HALLS in [self.comm, self.ref]:
                self.mc.communication.set_register(
                    self.PHASING_ACCURACY_REGISTER,
                    self.PHASING_ACCURACY_HALLS_DEFAULT,
                    servo=self.servo,
                    axis=self.axis,
                )
            else:
                self.mc.communication.set_register(
                    self.PHASING_ACCURACY_REGISTER,
                    self.PHASING_ACCURACY_DEFAULT,
                    servo=self.servo,
                    axis=self.axis,
                )

    @BaseTest.stoppable
    def setup(self) -> None:
        # Prerequisites:
        #  - Motor & Feedbacks configured (Pair poles & rated current are used)
        #  - Current control loop tuned
        #  - Feedbacks polarity checked
        # This test should be only executed for the reference sensor
        # Protection to avoid any unwanted movement
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.logger.info("CONFIGURATION OF THE TEST", axis=self.axis)

        self.comm = self.mc.configuration.get_commutation_feedback(servo=self.servo, axis=self.axis)
        self.ref = self.mc.configuration.get_reference_feedback(servo=self.servo, axis=self.axis)

        if self.ref == self.INTERNAL_GENERATOR_VALUE:
            raise TestError("Reference feedback sensor is set to internal generator")
        if self.comm == self.INTERNAL_GENERATOR_VALUE:
            raise TestError("Commutation feedback sensor is set to internal generator")

        # selection of commutation sensor
        if (
            self.mc.configuration.get_reference_feedback_category(servo=self.servo, axis=self.axis)
            == SensorCategory.INCREMENTAL
        ):
            # Delete commutation feedback from backup registers list as commutation feedback is kept the same
            fb = self.comm
        else:
            # Need to restore commutation feedback as in commutation feedback is set the one in reference
            fb = self.ref

        # Check phasing registers mode
        self.logger.debug("Checking input data")
        self.check_input_data()

        # Set sinusoidal commutation modulation
        self.mc.configuration.set_commutation_mode(
            CommutationMode.SINUSOIDAL, servo=self.servo, axis=self.axis
        )
        self.logger.info("Commutation modulation set to Sinusoidal")

        self.mc.motion.set_operation_mode(OperationMode.CURRENT, servo=self.servo, axis=self.axis)
        self.logger.info("Mode of operation set to Current mode")

        self.mc.motion.set_current_quadrature(0, servo=self.servo, axis=self.axis)
        self.mc.motion.set_current_direct(0, servo=self.servo, axis=self.axis)
        self.logger.info("Target quadrature current set to zero", axis=self.axis)
        self.logger.info("Target direct current set to zero", axis=self.axis)

        self.mc.configuration.set_commutation_feedback(fb, servo=self.servo, axis=self.axis)
        self.logger.info("Reset phasing status by setting again commutation sensor")

        self.mc.configuration.set_phasing_mode(PhasingMode.FORCED, servo=self.servo, axis=self.axis)
        self.logger.info("Set phasing mode to Forced")

        self.logger.info("Set phasing current to %.2f A", self.pha_current)

        pha_timeout = self.mc.communication.get_register(
            self.PHASING_TIMEOUT_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(pha_timeout, int):
            raise TypeError(f"{self.PHASING_TIMEOUT_REGISTER} has to be a integer")
        self.pha_timeout = pha_timeout
        self.logger.info(f"Set phasing timeout to {self.pha_timeout / 1000} s")

        pha_accuracy = self.mc.communication.get_register(
            self.PHASING_ACCURACY_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(pha_accuracy, int):
            raise TypeError(f"{self.PHASING_ACCURACY_REGISTER} has to be a integer")
        self.pha_accuracy = pha_accuracy
        self.logger.info(f"Set phasing accuracy to {self.pha_accuracy} mº")

        self.reaction_codes_to_warning()

    @BaseTest.stoppable
    def reaction_codes_to_warning(self) -> None:
        try:
            self.mc.communication.set_register(
                "COMMU_ANGLE_INTEGRITY1_OPTION", 1, servo=self.servo, axis=self.axis
            )
        except IMRegisterNotExist:
            self.logger.warning("Could not write COMMU_ANGLE_INTEGRITY1_OPTION")

        try:
            self.mc.communication.set_register(
                "COMMU_ANGLE_INTEGRITY2_OPTION", 1, servo=self.servo, axis=self.axis
            )
        except IMRegisterNotExist:
            self.logger.warning("Could not write COMMU_ANGLE_INTEGRITY2_OPTION")

    @BaseTest.stoppable
    def define_phasing_steps(self) -> int:
        # Doc: Last step is defined as the first angle delta smaller than 3 times the phasing accuracy
        delta = 3 * self.pha_accuracy / 1000

        # If reference feedback are Halls
        if self.ref == SensorType.HALLS:
            actual_angle = self.INITIAL_ANGLE_HALLS
        else:
            actual_angle = self.INITIAL_ANGLE
        num_of_steps = 1
        while actual_angle > delta:
            actual_angle = actual_angle / 2
            num_of_steps += 1
        return num_of_steps

    @BaseTest.stoppable
    def set_phasing_mode(self) -> PhasingMode:
        ref_category = self.mc.configuration.get_reference_feedback_category(
            servo=self.servo, axis=self.axis
        )
        comm_category = self.mc.configuration.get_commutation_feedback_category(
            servo=self.servo, axis=self.axis
        )
        # Check if reference feedback is incremental
        if ref_category == SensorCategory.INCREMENTAL:
            # In that if commutation feedback is incremental also
            if self.comm == SensorType.INTGEN:
                raise TestError("Commutation feedback sensor is set to internal generator")
            elif comm_category == SensorCategory.INCREMENTAL:
                # Set forced mode
                return PhasingMode.FORCED
            else:
                # Set a forced and then a No-Phasing
                return PhasingMode.NO_PHASING
        elif self.comm == self.ref:
            # Set a forced and then a No-Phasing
            return PhasingMode.NO_PHASING
        else:
            # Set a forced and then a Non forced
            return PhasingMode.NON_FORCED

    @BaseTest.stoppable
    def loop(self) -> ResultType:
        self.logger.info("START OF THE TEST", axis=self.axis)

        num_of_steps = self.define_phasing_steps()

        self.logger.info("Enabling motor", axis=self.axis)
        self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)

        self.logger.info("Wait until phasing is executed", axis=self.axis)

        phasing_bit_latched = False
        timeout = time.time() + (num_of_steps * self.pha_timeout / 1000)
        while time.time() < timeout:
            self.stoppable_sleep(0.1)
            if self.mc.configuration.is_commutation_feedback_aligned(
                servo=self.servo, axis=self.axis
            ):
                phasing_bit_latched = True
                break

        if not phasing_bit_latched:
            self.logger.info("ERROR: Phasing process has failed. Review your motor configuration.")
            return self.ResultType.FAIL

        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        pha = self.set_phasing_mode()
        ang = self.mc.communication.get_register(
            self.COMMUTATION_ANGLE_OFFSET_REGISTER, servo=self.servo, axis=self.axis
        )

        self.suggested_registers = {
            "COMMU_PHASING_MODE": pha,
            "COMMU_PHASING_TIMEOUT": self.pha_timeout,
            "COMMU_PHASING_ACCURACY": self.pha_accuracy,
            "COMMU_PHASING_MAX_CURRENT": self.pha_current,
            "COMMU_ANGLE_OFFSET": ang,
        }
        return self.ResultType.SUCCESS

    def teardown(self) -> None:
        self.logger.info("Disabling motor")
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)

    def get_result_msg(self, output: ResultType) -> str:
        return self.result_description[output]

    def get_result_severity(self, output: ResultType) -> SeverityLevel:
        if output < self.ResultType.SUCCESS:
            return SeverityLevel.FAIL
        else:
            return SeverityLevel.SUCCESS
